"""Views for the chat app."""

from rest_framework import permissions
from rest_framework.generics import ListAPIView, CreateAPIView, RetrieveUpdateDestroyAPIView, get_object_or_404

from aparnik.contrib.messaging.utils import notify

from .serializers import *


# chat session
class ChatSessionListAPIView(ListAPIView):
    serializer_class = ChatSessionListSerializer
    permission_classes = [permissions.IsAuthenticated]
    filter_class = []

    def get_queryset(self):
        user = self.request.user
        queryset = ChatSession.objects.user_session(user=user)
        return queryset


class ChatSessionDetailAPIView(RetrieveUpdateDestroyAPIView):
    serializer_class = ChatSessionDetailsSerializer
    queryset = ChatSession.objects.all()
    permission_classes = [permissions.IsAuthenticated]
    lookup_url_kwarg = 'uri'
    lookup_field = 'uri'


class ChatSessionCreateAPIView(CreateAPIView):
    serializer_class = ChatSessionDetailsSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        return ChatSession.objects.all()

    def perform_create(self, serializer):
        user = self.request.user
        serializer.save(owner=user)


# chat member
class ChatSessionMemberListAPIView(ListAPIView):
    serializer_class = ChatSessionMemberListSerializer
    permission_classes = [permissions.IsAuthenticated]
    filter_class = []

    def get_queryset(self):
        user = self.request.user
        uri = None
        if 'uri' in self.request.parser_context['kwargs']:
            uri = self.request.parser_context['kwargs']['uri']
        return ChatSessionMember.objects.filter(chat_session__uri=uri)


class ChatSessionMemberDetailAPIView(RetrieveUpdateDestroyAPIView):
    serializer_class = ChatSessionMemberDetailsSerializer
    queryset = ChatSessionMember.objects.all()
    permission_classes = [permissions.IsAuthenticated]
    lookup_url_kwarg = 'id'
    lookup_field = 'id'
    # TODO: check permission for delete id


# message list
class ChatSessionMessageListAPIView(ListAPIView):
    serializer_class = ChatSessionMessageListSerializer
    permission_classes = [permissions.IsAuthenticated]
    filter_class = []

    def get_queryset(self):
        user = self.request.user
        uri = None
        if 'uri' in self.request.parser_context['kwargs']:
            uri = self.request.parser_context['kwargs']['uri']
        return ChatSessionMessage.objects.filter(chat_session__uri=uri).order_by('-created_at')


class ChatSessionMessageDetailAPIView(RetrieveUpdateDestroyAPIView):
    serializer_class = ChatSessionMessageDetailsSerializer
    queryset = ChatSessionMessage.objects.all()
    permission_classes = [permissions.IsAuthenticated]
    lookup_url_kwarg = 'id'
    lookup_field = 'id'
    # TODO: check permission for delete id


class ChatSessionMessageCreateAPIView(CreateAPIView):
    serializer_class = ChatSessionMessageDetailsSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        return ChatSession.objects.all()

    def perform_create(self, serializer):
        user = self.request.user
        chat_session = None
        if 'uri' in self.request.parser_context['kwargs']:
            uri = self.request.parser_context['kwargs']['uri']
            chat_session = get_object_or_404(ChatSession.objects.all(), uri=uri)
        serializer.save(user=user, chat_session=chat_session)
        notify(
            uri=chat_session.uri,
            source=user,
            source_display_name=user.get_full_name(),
            recipient=None,
            action='Create',
            obj=serializer.instance,
            short_description='You a new message',
            extra_data={
                'uri': chat_session.uri,
                'message': serializer.data,
            },
            channels=['websocket'],
            silent=True,
        )
##
# class ChatSessionMessageView(APIView):
#     """Create/Get Chat session messages."""
#
#     permission_classes = (permissions.IsAuthenticated,)
#
#     def get(self, request, *args, **kwargs):
#         """return all messages in a chat session."""
#         uri = kwargs['uri']
#
#         chat_session = ChatSession.objects.get(uri=uri)
#         messages = [chat_session_message.to_json()
#                     for chat_session_message in chat_session.messages.all()]
#
#         return Response({
#             'id': chat_session.id, 'uri': chat_session.uri,
#             'messages': messages
#         })
#
#     def post(self, request, *args, **kwargs):
#         """create a new message in a chat session."""
#         uri = kwargs['uri']
#         message = request.data['message']
#
#         user = request.user
#         chat_session = ChatSession.objects.get(uri=uri)
#
#         chat_session_message = ChatSessionMessage.objects.create(
#             user=user, chat_session=chat_session, message=message
#         )
#
#         notif_args = {
#             'source': user,
#             'source_display_name': user.get_full_name(),
#             'category': 'chat', 'action': 'Sent',
#             'obj': chat_session_message.id,
#             'short_description': 'You a new message', 'silent': True,
#             'extra_data': {
#                 'uri': chat_session.uri,
#                 'message': chat_session_message.to_json()
#             }
#         }
#         notify.send(
#             sender=self.__class__, **notif_args, channels=['websocket']
#         )
#
#         return Response({
#             'status': 'SUCCESS', 'uri': chat_session.uri, 'message': message,
#             'user': deserialize_user(user)
#         })



"""Models for the chat app."""

from uuid import uuid4

from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from aparnik.contrib.filefields.models import FileField
from django.db import models
from django.contrib.auth import get_user_model

User = get_user_model()


class TrackableDateModel(models.Model):
    """Abstract model to Track the creation/updated date for a model."""

    created_at = models.DateTimeField(auto_now_add=True)
    update_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

    @property
    def resourcetype(self):
        return self._meta.object_name


def _generate_unique_uri():
    """Generates a unique uri for the chat session."""
    return str(uuid4()).replace('-', '')[:15]

class ChatSessionManager(models.Manager):
    def get_queryset(self):
        return super(ChatSessionManager, self).get_queryset()

    def user_session(self, user):
        queryset = self.filter(
            pk__in=ChatSessionMember.objects.filter(user=user).values_list('chat_session', flat=True).distinct())
        queryset = queryset | self.filter(owner=user)
        return queryset.distinct()


class ChatSession(TrackableDateModel):
    """
    A Chat Session.

    The uri's are generated by taking the first 15 characters from a UUID
    """

    owner = models.ForeignKey(User, on_delete=models.PROTECT)
    uri = models.URLField(default=_generate_unique_uri)

    objects = ChatSessionManager()

    def __str__(self):
        return self.uri

    def get_api_uri(self):
        return reverse('aparnik-api:chats:details', args=[self.uri])

    def get_member_api_uri(self):
        return reverse('aparnik-api:chats:members', args=[self.uri])

    def get_messages_api_uri(self):
        return reverse('aparnik-api:chats:messages', args=[self.uri])

    def get_messages_create_api_uri(self):
        return reverse('aparnik-api:chats:messages-create', args=[self.uri])


class ChatSessionMessage(TrackableDateModel):
    """Store messages for a session."""

    user = models.ForeignKey(User, on_delete=models.PROTECT)
    chat_session = models.ForeignKey(
        ChatSession, related_name='messages', on_delete=models.PROTECT
    )
    message = models.TextField(default='', max_length=2000)
    file_obj= models.ForeignKey(FileField, null=True, blank=True, on_delete=models.CASCADE, verbose_name=_('File'))

    def get_api_uri(self):
        return reverse('aparnik-api:chats:messages-details', args=[self.pk])


class ChatSessionMember(TrackableDateModel):
    """Store all users in a chat session."""

    chat_session = models.ForeignKey(
        ChatSession, related_name='members', on_delete=models.PROTECT
    )
    user = models.ForeignKey(User, on_delete=models.PROTECT)

    def get_api_uri(self):
        return reverse('aparnik-api:chats:members-details', args=[self.pk])



"""Base Implementation of a Delivery Backend."""

import abc
from json import dumps
import six
import channels.layers
from asgiref.sync import async_to_sync

from django.contrib.auth import get_user_model

from . import default_settings as settings


@six.add_metaclass(abc.ABCMeta)
class BaseNotificationChannel():
    """Base channel for sending notifications."""

    def __init__(self, **kwargs):
        self.notification_kwargs = kwargs

    @abc.abstractmethod
    def construct_message(self):
        """Constructs a message from notification details."""
        pass

    @abc.abstractmethod
    def notify(self, message):
        """Sends the notification."""
        pass


class ConsoleChannel(BaseNotificationChannel):
    """Dummy channel that prints to the console."""

    def construct_message(self):
        """Stringify the notification kwargs."""
        return str(self.notification_kwargs)

    def notify(self, message):
        print(message)


class BasicWebSocketChannel(BaseNotificationChannel):
    """It creates a Redis user for each user (based on their username)."""

    def _connect(self):

        channel_layer = channels.layers.get_channel_layer()
        return channel_layer

    def construct_message(self):
        """Construct message from notification details."""

        return self.notification_kwargs

    def notify(self, message):
        """
        Puts a new message on the queue.
        
        The queue is named based on the username (for Uniqueness)
        """
        uri = self.notification_kwargs['uri']

        if not uri:
            return

        channel_layer = self._connect()
        # add this moment work only with user, not guest
        async_to_sync(channel_layer.group_send)(
            uri,
            {
                "type": "send_data",
                'data': message,
            }
        )

       # # Get user instance
        # User = get_user_model()
        # source = User.objects.get(id=message['source'])
        # recipient = User.objects.get(id=message['recipient'])
        #
        # channel.queue_declare(queue=source.username)
        #
        # jsonified_messasge = dumps(message)
        # channel.basic_publish(
        #     exchange='', routing_key=recipient.username,
        #     body=jsonified_messasge
        # )
